local goHandle, goEngine                = ...
local goPlugin                          = LibStub('AceAddon-3.0'):NewAddon(goHandle, 'AceConsole-3.0', 'AceEvent-3.0', 'AceTimer-3.0')
local L                                 = LibStub('AceLocale-3.0'):GetLocale(goHandle)
local goDb

local goOptions
local goDefault                         = {                                     -- default database settings
  global                                = {
    defaults                            = {
      isEnabled                         = true,
      forceLoot                         = true,
      lootDelay                         = 0,
      lootCount                         = 50,
      hidePopup                         = false
    }
  }
}

goEngine[1]                             = goPlugin
goEngine[2]                             = goHandle
_G.goLoot                               = goEngine

local ChatEdit_SendText                 = ChatEdit_SendText
local GetContainerItemInfo              = GetContainerItemInfo
local GetContainerNumFreeSlots          = GetContainerNumFreeSlots
local GetContainerNumSlots              = GetContainerNumSlots
local GetCVar                           = GetCVar
local GetItemInfo                       = GetItemInfo
local GetLootInfo                       = GetLootInfo
local GetLootMethod                     = GetLootMethod
local GetLootSlotType                   = GetLootSlotType
local GetLootThreshold                  = GetLootThreshold
local GetNumLootItems                   = GetNumLootItems
local GetNumQuestLeaderBoards           = GetNumQuestLeaderBoards
local GetNumQuestLogEntries             = C_QuestLog.GetNumQuestLogEntries
local GetQuestLogInfo                   = C_QuestLog.GetInfo
local InCombatLockdown                  = InCombatLockdown
local IsAddOnLoaded                     = IsAddOnLoaded
local IsInGroup                         = IsInGroup
local LootSlot                          = LootSlot
local LootSlotHasItem                   = LootSlotHasItem
local PlaySoundFile                     = PlaySoundFile
local SetCVar                           = SetCVar
local StaticPopup_Show                  = StaticPopup_Show
local UnitClass                         = UnitClass
local UnitRace                          = UnitRace
local UnitSex                           = UnitSex

local LOOTTYPE_CURRENCY                 = {LOOT_SLOT_MONEY, LOOT_SLOT_CURRENCY}

-- worgen has sound files for human and worgen forms
-- demon hunter has unique sound files
local INVSOUND_FULL                     = {
  [0]                                   = { -- default
    [1]  = {[2] =  540752, [3] =  540575},  -- human
    [2]  = {[2] =  541273, [3] =  541162},  -- orc
    [3]  = {[2] =  539891, [3] =  539832},  -- dwarf
    [4]  = {[2] =  540960, [3] =  540907},  -- night elf
    [5]  = {[2] =  542631, [3] =  542538},  -- undead
    [6]  = {[2] =  542922, [3] =  542857},  -- tauren
    [7]  = {[2] =  540365, [3] =  540324},  -- gnome
    [8]  = {[2] =  543169, [3] =  543100},  -- troll
    [9]  = {[2] =  541968, [3] =  541783},  -- goblin
    [10] = {[2] =  539183, [3] =  539184},  -- blood elf
    [11] = {[2] =  539479, [3] =  539550},  -- draenei
    [22] = {[2] =  541641, [3] =  541462},  -- worgen [human]
    [24] = {[2] =  630154, [3] =  636483},  -- pandaren neutral
    [25] = {[2] =  630154, [3] =  636483},  -- pandaren alliance
    [26] = {[2] =  630154, [3] =  636483},  -- pandaren horde
    [27] = {[2] = 1732273, [3] = 1731899},  -- nightborne
    [28] = {[2] = 1730777, [3] = 1730403},  -- highmountain tauren
    [29] = {[2] = 1733030, [3] = 1732652},  -- void elf
    [30] = {[2] = 1731525, [3] = 1731899},  -- lightforged draenei
    [31] = {[2] = 1903455, [3] = 1902942},  -- zandalari troll
    [32] = {[2] = 2492017, [3] = 2531005},  -- kul tiran
    [34] = {[2] = 1902429, [3] = 1901916},  -- dark iron dwarf
    [35] = {[2] = 3106619, [3] = 3106154},  -- vulpera
    [36] = {[2] = 1950899, [3] = 1950898},  -- mag'har orc
    [37] = {[2] = 3107084, [3] = 3107553}   -- mechagnome
  },
  [1]                                   = { -- demon hunter
    [4]  = {[2] = 1304825, [3] = 1304528},  -- night elf
    [10] = {[2] = 1306421, [3] = 1284686}   -- blood elf
  },
  [2]                                   = { -- worgen
    [22] = {[2] =  542170, [3] =  542054}   -- worgen [beast]
  }
}

StaticPopupDialogs['GL_POPUP_FORCELOOT_ENABLED'] = {
  text = L['GL_POPUP_FORCELOOT_ENABLED'],
  button1 = L['GL_POPUP_YES'],
  button3 = L['GL_POPUP_IGNORE'],
  button2 = L['GL_POPUP_NO'],
  OnAccept = function()
    SetCVar('autoLootDefault', 0)
  end,
  OnAlt = function()
    db.hidePopup = true
  end,
  timeout = 0,
  whileDead = true,
  hideOnEscape = false,
  preferredIndex = 3
}

StaticPopupDialogs['GL_POPUP_FORCELOOT_DISABLED'] = {
  text = L['GL_POPUP_FORCELOOT_DISABLED'],
  button1 = L['GL_POPUP_YES'],
  button3 = L['GL_POPUP_IGNORE'],
  button2 = L['GL_POPUP_NO'],
  OnAccept = function()
    SetCVar('autoLootDefault', 1)
  end,
  OnAlt = function()
    db.hidePopup = true
  end,
  timeout = 0,
  whileDead = true,
  hideOnEscape = false,
  preferredIndex = 3
}

table.contains = function(table, search)
  for _, v in ipairs(table) do
    if v == search then
      return true
    end
  end

  return false
end

function goPlugin:OnInitialize()
  self.db = LibStub('AceDB-3.0'):New(goHandle .. 'DB', goDefault)                       -- register the database
  goDb = self.db.global.defaults

  LibStub('AceConfigRegistry-3.0'):RegisterOptionsTable(goHandle, goOptions)    -- register the options table
  self.blizzPanel = LibStub('AceConfigDialog-3.0'):AddToBlizOptions(goHandle, L['GL'])

  self:RegisterChatCommand(L['GL_SLASHCMD'], function()                          -- register the slash command
    if InCombatLockdown() then
      goPlugin:Print(L['GL_INCOMBATLOCKDOWN'])
      return
    end

    InterfaceOptionsFrame:Show()
    InterfaceOptionsFrame_OpenToCategory(goHandle)
  end)

  self:SetLootState(false)                                                      -- set the default loot state to false
  self:RegisterEvent('LOOT_READY', 'OnLootOpened')                              -- register the events and scripts
  self:RegisterEvent('LOOT_CLOSED', 'OnLootClosed')

  InterfaceOptionsFrame:HookScript('OnHide', function()                         -- ensure that the optimal settings is used
    goPlugin:ShowSettingPopup()
  end)
  self:ScheduleTimer(function()
    goPlugin:ShowSettingPopup()
  end, 5)

end

local itemBag, soundIndex
function goPlugin:IsItemStorable(itemName, itemCount)
  function getItemInfo(bId, sId)
    itemBag = select(7, GetContainerItemInfo(bId, sId))
    return {
      ['name']  = select(1, GetItemInfo(itemBag)),
      ['count'] = select(2, GetContainerItemInfo(bId, sId)),
      ['stack'] = select(8, GetItemInfo(itemBag))
    }
  end

  for bId = 0, 4 do
    if GetContainerNumFreeSlots(bId) > 0 then
      return true
    end
  end

  for bId = 0, 4 do
    for sId = 1, GetContainerNumSlots(bId) do
      itemBag = getItemInfo(bId, sId)

      if itemBag.name == itemName and (itemBag.stack - itemBag.count) >= itemCount then
        return true
      end
    end
  end

  if not self.lootSound and GetCVar('autoLootDefault') == '0' then              -- plays inventory full sound when Auto Loot is turned off
    PlaySoundFile(INVSOUND_FULL[(select(3, UnitRace('player')) == 22 and not select(2, HasAlternateForm())) and 2 or (select(3, UnitClass('player')) == 12 and 1 or 0)][select(3, UnitRace('player'))][UnitSex('player')])
    self.lootSound = true
  end
  return false
end

local numObjectives, infoQuest, itemQuest
function goPlugin:IsItemQuestCompleted(itemName)
  for qId = 1, GetNumQuestLogEntries() do
    numObjectives = GetQuestLogInfo(qId).isHeader and 0 or GetNumQuestLeaderBoards(qId)

    if numObjectives > 0 then
      for bId = 1, numObjectives do
        infoQuest = {GetQuestLogLeaderBoard(bId, qId)}
        itemQuest = select(3, infoQuest[1]:match('(%d)/(%d) (.*)'))

        if infoQuest[2] ~= nil and (infoQuest[2] == 'item' or infoQuest[2] == 'object') and itemQuest == itemName then
          return infoQuest[3]
        end
      end
    end
  end

  return false
end

local itemLoot
function goPlugin:LootSlot(itemSlot)
  itemLoot = self.lootTable[itemSlot]
  if not LootSlotHasItem(itemSlot) or itemLoot == nil then
    return
  end

  if table.contains(LOOTTYPE_CURRENCY, GetLootSlotType(itemSlot)) then
    LootSlot(itemSlot)
    return
  end

  itemLoot.locked = itemLoot.locked and true or (itemLoot.isQuestItem and self:IsItemQuestCompleted(itemLoot.item) or false)
  if itemLoot.locked or (self.isGrouped and itemLoot.quality >= GetLootThreshold()) then
    return
  end

  if itemLoot.isQuestItem or self:IsItemStorable(itemLoot.item, itemLoot.quantity) then
    LootSlot(itemSlot)
  end
end

function goPlugin:OnLootClosed()
  if not goDb.isEnabled or not self.isLooting then
    self:SetLootState(false)
    return
  end

  if IsAddOnLoaded('TradeSkillMaster') and TSMDestroyBtn:IsVisible() then       -- work around for TSM Destroy bugging out with goLoot
    TSMDestroyBtn:GetParent():GetParent():Hide()
    self:ScheduleTimer(function()
      if not goPlugin.editFrame then
        goPlugin.editFrame = CreateFrame('EditBox')
      end

      goPlugin.editFrame:Show()
      goPlugin.editFrame:SetText('/tsm destroy')
  	  ChatEdit_SendText(goPlugin.editFrame)
      goPlugin.editFrame:Hide()
    end, 0)
  end

  self:SetLootState(false)
end

function goPlugin:OnLootOpened(_, autoLoot)
  autoLoot = goDb.forceLoot and true or autoLoot

  if not goDb.isEnabled or not autoLoot or self.isLooting then
    self:SetLootState(false)
    return
  end

  self:SetLootState(true)
end

function goPlugin:OnLootUpdate()
  if not goDb.isEnabled or not self.isLooting then
    self:SetLootState(false)
    return
  end

  self.lootCount = self.lootCount - 1
  for lootSlot = GetNumLootItems(), 1, -1 do
    self:LootSlot(lootSlot)
  end

  if self.lootCount == 0 then
    self:SetLootState(false)
  end
end

function goPlugin:SetLootState(isLooting)
  self.isLooting = isLooting
  self.isGrouped = isLooting and (IsInGroup() and GetLootMethod() == 'master') or false
  self.lootCount = isLooting and goDb.lootCount or 0
  self.lootTable = isLooting and GetLootInfo() or {}
  self.lootSound = false
  self:CancelTimer(self.lootTimer)

  if isLooting then
    self.lootTimer = self:ScheduleRepeatingTimer('OnLootUpdate', (goDb.lootDelay / 1000))
  end
end

function goPlugin:ShowSettingPopup()
  if goDb.hidePopup then
    return
  end

  if goDb.isEnabled and goDb.forceLoot and GetCVar('autoLootDefault') ~= '0' then
    StaticPopup_Show('GL_POPUP_FORCELOOT_ENABLED')
  end

  if goDb.isEnabled and not goDb.forceLoot and GetCVar('autoLootDefault') ~= '1' then
    StaticPopup_Show('GL_POPUP_FORCELOOT_DISABLED')
  end
end

goOptions                               = {
  type                                  = 'group',
  name                                  = '',
  args                                  = {
    descHeader                          = {
      type                              = 'description',
      name                              = L['GL_OPTION_HEADER'],
      fontSize                          = 'medium',
      order                             = 1
    },
    isEnabled                           = {
      type                              = 'toggle',
      name                              = L['GL_OPTION_ENABLED'],
      desc                              = L['GL_OPTION_ENABLED_DESC'],
      get                               = function() return goDb.isEnabled end,
      set                               = function(_, value) goDb.isEnabled = value end,
      order                             = 2
    },
    descSpace01                         = {
      type                              = 'description',
      name                              = '',
      order                             = 3
    },
    forceLoot                           = {
      type                              = 'toggle',
      name                              = L['GL_OPTION_FORCELOOT'],
      desc                              = L['GL_OPTION_FORCELOOT_DESC'],
      disabled                		      = function() return not goDb.isEnabled end,
      get                               = function() return goDb.forceLoot end,
      set                               = function(_, value) goDb.forceLoot = value end,
      order                             = 4
    },
    lootDelay                           = {
      type                    		      = 'range',
      name                    		      = L['GL_OPTION_LOOTDELAY'],
      desc                    		      = L['GL_OPTION_LOOTDELAY_DESC'],
      disabled                		      = function() return not goDb.isEnabled end,
      get                     		      = function() return goDb.lootDelay end,
      set                     		      = function(_, value) goDb.lootDelay = value end,
      min                     		      = 0,
      max                     		      = 250,
      step                    		      = 10,
      order                   		      = 5
    },
    lootCount                           = {
      type                    		      = 'range',
      name                    		      = L['GL_OPTION_LOOTCOUNT'],
      desc                    		      = L['GL_OPTION_LOOTCOUNT_DESC'],
      disabled                		      = function() return not goDb.isEnabled end,
      get                     		      = function() return goDb.lootCount end,
      set                     		      = function(_, value) goDb.lootCount = value end,
      min                     		      = 10,
      max                     		      = 500,
      step                    		      = 10,
      order                   		      = 6
    },
    hidePopup                           = {
      type                              = 'toggle',
      name                              = L['GL_OPTION_HIDEPOPUP'],
      desc                              = L['GL_OPTION_HIDEPOPUP_DESC'],
      disabled                		      = function() return not goDb.isEnabled end,
      get                               = function() return goDb.hidePopup end,
      set                               = function(_, value) goDb.hidePopup = value end,
      order                             = 7
    }
  },
  order                                 = 1
}
